
import operator

import numpy as np
import matplotlib.pyplot as plt

from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_squared_error, r2_score
from sklearn.preprocessing import PolynomialFeatures


def generate_data():
    np.random.seed(0)
    n_data = 100
    stdev = 20
    # generate random data points (x,y)
    x = 2 - 3 * np.random.normal(0, 1, n_data)
    y = x - 2 * (x ** 2) + 0.5 * (x ** 3) + 0.01 * (x ** 4) + \
        0.001 * (x ** 5) + np.random.normal(-stdev, stdev, n_data)

    # transforming the data to include another axis
    x = x[:, np.newaxis]
    y = y[:, np.newaxis]
    return x, y


def fit_model(x, y, degree):
    # create the polynomal features to be used for training the regression model
    polynomial_features = PolynomialFeatures(degree=degree)
    x_poly = polynomial_features.fit_transform(x)
    # create a linear regression model using scikit-learn
    model = LinearRegression()
    # fit the model to the data points (x,y)
    model.fit(x_poly, y)
    # use the model to predict the data points
    y_poly_pred = model.predict(x_poly)
    # get the prediction error as RMSE
    rmse = np.sqrt(mean_squared_error(y, y_poly_pred))
    # get the regression score
    r2 = r2_score(y, y_poly_pred)
    return x_poly, y_poly_pred, rmse, r2


def plot_line(x, y_poly_pred):
    # sort the values of x before line plot
    sort_axis = operator.itemgetter(0)
    sorted_zip = sorted(zip(x, y_poly_pred), key=sort_axis)
    x, y_poly_pred = zip(*sorted_zip)
    plt.plot(x, y_poly_pred)


def show_legend(degrees):
    plt.legend([str(d) for d in degrees])


# main program
if __name__ == "__main__":
    # generate random(ish) data
    x, y = generate_data()
    # plot the data points (x,y)
    plt.scatter(x, y, s=20)

    # TODO 1: experiment with different range of polynomial orders
    degrees = range(1, 10)
    rmse_vect = []
    r2_vect = []
    # for each polynomial degree, create a regression model and plot the predicted outputs
    for deg in degrees:
        _, y_poly_pred, rmse1, r2 = fit_model(x, y, deg)
        # save the prediction score (RMSE, R2)
        rmse_vect.append(rmse1)
        r2_vect.append(r2)
        # plot the predicted outputs
        plot_line(x, y_poly_pred)

    show_legend(degrees)
    plt.title("Polynomial regression")
    plt.xlabel("x")
    plt.ylabel("y")
    plt.show()

    # Tip: you can close the first figure to show the second one
    print(degrees)
    print(rmse_vect)

    plt.plot(degrees, rmse_vect)
    plt.title("RMSE evaluation")
    plt.xlabel("model order")
    plt.ylabel("RMSE")
    plt.show()
    plt.plot(degrees, r2_vect)
    plt.title("Regression model fit")
    plt.xlabel("model order")
    plt.ylabel("R2 score")
    plt.show()

    # TODO 1: experiment with different range of polynomial orders
    # TODO 2: based on the experimental results, answer the following questions:
    # Q1: Which is the optimal polynomial order for this dataset with regards to the RMSE and model complexity. Tip: the RMSE improvement starts to decrease after a certain polynomial order on the RMSE chart.
    # Q2: Which is the optimal polynomial order for this dataset with regards to the R2 coefficient and model complexity. Tip: the R2 coefficient improvement starts to decrease after a certain polynomial order on the R2 chart.
    # Q3: Explain the results based on the provided function that is used to generate the dataset.
